@@ -1,4 +1,4 @@
-/* CARDPCH.C    (c) Copyright Roger Bowler, 1999                     */
+/* CARDPCH.C    (c) Copyright Roger Bowler, 1999-2001                */
 /*              ESA/390 Card Punch Device Handler                    */
 
 /*-------------------------------------------------------------------*/
@@ -155,9 +155,9 @@ BYTE            c;                      /* Output character          */
     {
         rc = open (dev->filename,
 #ifdef WIN32
-                    O_WRONLY | O_CREAT | O_TRUNC | O_SYNC | O_BINARY,
+                    O_WRONLY | O_CREAT | O_TRUNC /* | O_SYNC */ |  O_BINARY,
 #else /* WIN32 */
-                    O_WRONLY | O_CREAT | O_TRUNC | O_SYNC,
+                    O_WRONLY | O_CREAT | O_TRUNC /* | O_SYNC */,
 #endif /* WIN32 */
                     S_IRUSR | S_IWUSR | S_IRGRP);
         if (rc < 0)
@@ -227,7 +227,7 @@ BYTE            c;                      /* Output character          */
             else
             {
                 /* Pad card image with blanks */
-                for (i = dev->cardpos; i < CARD_LENGTH; i++);
+                for (i = dev->cardpos; i < CARD_LENGTH; i++)
                     dev->buf[i] = HEX40;
             }
 
