@@ -1,4 +1,4 @@
-/* DASDINIT.C   (c) Copyright Roger Bowler, 1999-2000                */
+/* DASDINIT.C   (c) Copyright Roger Bowler, 1999-2001                */
 /*              Hercules DASD Utilities: DASD image builder          */
 
 /*-------------------------------------------------------------------*/
@@ -12,7 +12,7 @@
 /* filename     is the name of the disk image file to be created     */
 /*              (this program will not overwrite an existing file)   */
 /* devtype      is the emulated device type.                         */
-/*              CKD: 2311, 2314, 3330, 3350, 3380, 3390              */
+/*              CKD: 2311, 2314, 3330, 3350, 3380, 3390, 9345        */
 /*              FBA: 3310, 3370                                      */
 /* volser       is the volume serial number (1-6 characters)         */
 /* size         is the size of the device (in cylinders for CKD      */
@@ -30,6 +30,11 @@ BYTE iplpsw[8]    = {0x00, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0F};
 BYTE iplccw1[8]   = {0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01};
 BYTE iplccw2[8]   = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
 
+#ifdef EXTERNALGUI
+/* Special flag to indicate whether or not we're being
+   run under the control of the external GUI facility. */
+int  extgui = 0;
+#endif /*EXTERNALGUI*/
 
 /*-------------------------------------------------------------------*/
 /* Subroutine to display command syntax and exit                     */
@@ -41,7 +46,7 @@ argexit ( int code )
             "Syntax:\tdasdinit filename devtype volser size\n"
             "where:\tfilename = name of file to be created\n"
             "\tdevtype  = 2311, 2314, 3330, 3340, 3350, 3375, 3380, "
-            "3390 (CKD)\n"
+            "3390, 9345(CKD)\n"
             "\t           3310, 3370 (FBA)\n"
             "\tvolser   = volume serial number (1-6 characters)\n"
             "\tsize     = volume size in cylinders (CKD devices)\n"
@@ -136,7 +141,14 @@ int             highcyl;                /* CKD header high cyl number*/
     {
         /* Display progress message every 10 cylinders */
         if ((cyl % 10) == 0)
+#ifdef EXTERNALGUI
+        {
+            if (extgui) fprintf (stderr, "CYL=%u\n", cyl);
+            else fprintf (stderr, "Writing cylinder %u\r", cyl);
+        }
+#else /*!EXTERNALGUI*/
             fprintf (stderr, "Writing cylinder %u\r", cyl);
+#endif /*EXTERNALGUI*/
 
         for (head = 0; head < heads; head++)
         {
@@ -452,7 +464,14 @@ U32             maxsect;                /* Maximum sector count      */
 
         /* Display progress message every 100 sectors */
         if ((sectnum % 100) == 0)
+#ifdef EXTERNALGUI
+        {
+            if (extgui) fprintf (stderr, "BLK=%u\n", sectnum);
+            else fprintf (stderr, "Writing sector %u\r", sectnum);
+        }
+#else /*!EXTERNALGUI*/
             fprintf (stderr, "Writing sector %u\r", sectnum);
+#endif /*EXTERNALGUI*/
 
         /* Write the sector to the file */
         rc = write (fd, buf, sectsz);
@@ -500,12 +519,17 @@ BYTE    volser[7];                      /* Volume serial number      */
 BYTE    c;                              /* Character work area       */
 
     /* Display the program identification message */
-    fprintf (stderr,
-            "Hercules DASD image file creation program %s "
-            "(c)Copyright Roger Bowler, 1999\n",
-            MSTRING(VERSION));
+    display_version (stderr,
+                     "Hercules DASD image file creation program\n");
 
     /* Check the number of arguments */
+#ifdef EXTERNALGUI
+    if (argc >= 1 && strncmp(argv[argc-1],"EXTERNALGUI",11) == 0)
+    {
+        extgui = 1;
+        argc--;
+    }
+#endif /*EXTERNALGUI*/
     if (argc != 5)
         argexit(5);
 
@@ -585,6 +609,12 @@ BYTE    c;                              /* Character work area       */
         maxdlen = 56664;
         break;
 
+    case 0x9345:
+        type = 'C';
+        heads = 15;
+        maxdlen = 46456;
+        break;
+
     case 0x3310:
         type = 'F';
         sectsize = 512;
