@@ -1,4 +1,4 @@
-/* DASDISUP.C   (c) Copyright Roger Bowler, 1999                     */
+/* DASDISUP.C   (c) Copyright Roger Bowler, 1999-2001                */
 /*              Hercules DASD Utilities: IEHIOSUP                    */
 
 /*-------------------------------------------------------------------*/
@@ -74,6 +74,12 @@ static BYTE *secondload[] = {
         "IGG094",                       /* TSO (SVC94)               */
         NULL };                         /* End of list               */
 
+#ifdef EXTERNALGUI
+/* Special flag to indicate whether or not we're being
+   run under the control of the external GUI facility. */
+int  extgui = 0;
+#endif /*EXTERNALGUI*/
+
 #if 0
 /*-------------------------------------------------------------------*/
 /* Subroutine to process a member                                    */
@@ -605,12 +611,17 @@ MEMINFO        *memtab;                 /* -> Member info array      */
 int             nmem = 0;               /* Number of array entries   */
 
     /* Display the program identification message */
-    fprintf (stdout,
-            "Hercules IEHIOSUP program %s "
-            "(c)Copyright Roger Bowler, 1999\n",
-            MSTRING(VERSION));
+    display_version (stderr,
+                     "Hercules IEHIOSUP program ");
 
     /* Check the number of arguments */
+#ifdef EXTERNALGUI
+    if (argc >= 1 && strncmp(argv[argc-1],"EXTERNALGUI",11) == 0)
+    {
+        extgui = 1;
+        argc--;
+    }
+#endif /*EXTERNALGUI*/
     if (argc != 2)
     {
         fprintf (stdout,
@@ -682,9 +693,16 @@ int             nmem = 0;               /* Number of array entries   */
             "End of directory: %d members selected\n",
             nmem);
 
+#ifdef EXTERNALGUI
+    if (extgui) fprintf (stderr,"NMEM=%d\n",nmem);
+#endif /*EXTERNALGUI*/
+
     /* Read each member and resolve the embedded TTRs */
     for (i = 0; i < nmem; i++)
     {
+#ifdef EXTERNALGUI
+        if (extgui) fprintf (stderr,"MEM=%d\n",i);
+#endif /*EXTERNALGUI*/
         rc = resolve_xctltab (cif, noext, extent, memtab+i,
                                 memtab, nmem);
 
