@@ -1,4 +1,4 @@
-/* FBADASD.C    (c) Copyright Roger Bowler, 1999-2000                */
+/* FBADASD.C    (c) Copyright Roger Bowler, 1999-2001                */
 /*              ESA/390 FBA Direct Access Storage Device Handler     */
 
 /*-------------------------------------------------------------------*/
@@ -68,10 +68,14 @@ BYTE    c;                              /* Character work area       */
     dev->fd = open (dev->filename, O_RDWR|O_BINARY);
     if (dev->fd < 0)
     {
-        fprintf (stderr,
-                "HHC302I File %s open error: %s\n",
-                dev->filename, strerror(errno));
-        return -1;
+        dev->fd = open (dev->filename, O_RDONLY|O_BINARY);
+        if (dev->fd < 0)
+        {
+            fprintf (stderr,
+                    "HHC302I File %s open error: %s\n",
+                    dev->filename, strerror(errno));
+            return -1;
+        }
     }
 
     /* Determine the device size */
@@ -334,7 +338,7 @@ int     repcnt;                         /* Replication count         */
             /* Handle read error condition */
             if (rc < 0)
                 logmsg ("HHC312I Read error in file %s: %s\n",
-                        dev->filename, strerror(errno))
+                        dev->filename, strerror(errno));
             else
                 logmsg ("HHC313I Unexpected end of file in %s\n",
                         dev->filename);
@@ -494,7 +498,7 @@ int     repcnt;                         /* Replication count         */
                 /* Handle read error condition */
                 if (rc < 0)
                     logmsg ("HHC316I Read error in file %s: %s\n",
-                            dev->filename, strerror(errno))
+                            dev->filename, strerror(errno));
                 else
                     logmsg ("HHC317I Unexpected end of filein %s\n",
                             dev->filename);
@@ -922,7 +926,7 @@ int     blkfactor;                      /* Number of device blocks
             /* Handle read error condition */
             if (rc < 0)
                 logmsg ("HHC323I Read error in file %s: %s\n",
-                        dev->filename, strerror(errno))
+                        dev->filename, strerror(errno));
             else
                 logmsg ("HHC324I Unexpected end of file in %s\n",
                         dev->filename);
