@@ -1,7 +1,8 @@
-/* IO.C         (c) Copyright Roger Bowler, 1994-2000                */
+/* IO.C         (c) Copyright Roger Bowler, 1994-2001                */
 /*              ESA/390 CPU Emulator                                 */
 
-/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2000      */
+/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2001      */
+/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2001      */
 
 /*-------------------------------------------------------------------*/
 /* This module implements all I/O instructions of the                */
@@ -24,7 +25,9 @@
 /*      CONCS instruction added - Jan Jaeger                         */
 /*      DISCS instruction added - Jan Jaeger                         */
 /*      TPI fix - Jay Maynard, found by Greg Smith                   */
+/*      STCRW instruction nullification correction - Jan Jaeger      */
 /*      I/O rate counter - Valery Pogonchenko                        */
+/*      64-bit IDAW support - Roger Bowler v209                  @IWZ*/
 /*-------------------------------------------------------------------*/
 
 #include "hercules.h"
@@ -33,32 +36,29 @@
 
 #include "inline.h"
 
-
 #if defined(FEATURE_CHANNEL_SUBSYSTEM)
 
 /*-------------------------------------------------------------------*/
 /* B230 CSCH  - Clear Subchannel                                 [S] */
 /*-------------------------------------------------------------------*/
-void zz_clear_subchannel (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(clear_subchannel)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 DEVBLK *dev;                            /* -> device block           */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
     /* Program check if reg 1 bits 0-15 not X'0001' */
-    if ( (regs->gpr[1] >> 16) != 0x0001 )
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+    if ( regs->GR_LHH(1) != 0x0001 )
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
     /* Locate the device block for this subchannel */
-    dev = find_device_by_subchan (regs->gpr[1] & 0xFFFF);
+    dev = find_device_by_subchan (regs->GR_LHL(1));
 
     /* Condition code 3 if subchannel does not exist,
        is not valid, or is not enabled */
@@ -81,26 +81,24 @@ DEVBLK *dev;                            /* -> device block           */
 /*-------------------------------------------------------------------*/
 /* B231 HSCH  - Halt Subchannel                                  [S] */
 /*-------------------------------------------------------------------*/
-void zz_halt_subchannel (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(halt_subchannel)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 DEVBLK *dev;                            /* -> device block           */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
     /* Program check if reg 1 bits 0-15 not X'0001' */
-    if ( (regs->gpr[1] >> 16) != 0x0001 )
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+    if ( regs->GR_LHH(1) != 0x0001 )
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
     /* Locate the device block for this subchannel */
-    dev = find_device_by_subchan (regs->gpr[1] & 0xFFFF);
+    dev = find_device_by_subchan (regs->GR_LHL(1));
 
     /* Condition code 3 if subchannel does not exist,
        is not valid, or is not enabled */
@@ -121,17 +119,15 @@ DEVBLK *dev;                            /* -> device block           */
 /*-------------------------------------------------------------------*/
 /* B232 MSCH  - Modify Subchannel                                [S] */
 /*-------------------------------------------------------------------*/
-void zz_modify_subchannel (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(modify_subchannel)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 DEVBLK *dev;                            /* -> device block           */
 PMCW    pmcw;                           /* Path management ctl word  */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
@@ -139,21 +135,21 @@ PMCW    pmcw;                           /* Path management ctl word  */
     FW_CHECK(effective_addr2, regs);
 
     /* Fetch the updated path management control word */
-    vfetchc ( &pmcw, sizeof(PMCW)-1, effective_addr2, b2, regs );
+    ARCH_DEP(vfetchc) ( &pmcw, sizeof(PMCW)-1, effective_addr2, b2, regs );
 
     /* Program check if reserved bits are not zero */
     if (pmcw.flag4 & PMCW4_RESV
         || (pmcw.flag5 & PMCW5_LM) == PMCW5_LM_RESV
         || pmcw.flag24 != 0 || pmcw.flag25 != 0
         || pmcw.flag26 != 0 || (pmcw.flag27 & PMCW27_RESV))
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
     /* Program check if reg 1 bits 0-15 not X'0001' */
-    if ( (regs->gpr[1] >> 16) != 0x0001 )
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+    if ( regs->GR_LHH(1) != 0x0001 )
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
     /* Locate the device block for this subchannel */
-    dev = find_device_by_subchan (regs->gpr[1] & 0xFFFF);
+    dev = find_device_by_subchan (regs->GR_LHL(1));
 
     /* Condition code 3 if subchannel does not exist */
     if (dev == NULL)
@@ -226,53 +222,62 @@ PMCW    pmcw;                           /* Path management ctl word  */
 }
 
 
-#if defined(FEATURE_INTERPRETIVE_EXECUTION)
 /*-------------------------------------------------------------------*/
 /* B23B RCHP  - Reset Channel Path                               [S] */
 /*-------------------------------------------------------------------*/
-void zz_reset_channel_path (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(reset_channel_path)
 {
 int     b2;                             /* Base of effective addr    */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
+BYTE    chpid;
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    if(!regs->sie_state)
-        program_interrupt(regs, PGM_OPERATION_EXCEPTION);
-
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
+    /* Program check if reg 1 bits 0-23 not zero */
+    if(regs->GR_L(1) & 0xFFFFFF00)
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
+
+    chpid = effective_addr2 & 0xFF;
+
+    if( !(regs->psw.cc = chp_reset(chpid)) )
+    {
+        obtain_lock(&sysblk.intlock);
+        sysblk.chp_reset[chpid/32] |= 0x80000000 >> (chpid % 32);
+        ON_IC_CHANRPT;
+        WAKEUP_WAITING_CPU (ALL_CPUS, CPUSTATE_STARTED);
+        release_lock (&sysblk.intlock);
+    }
+
+    RETURN_INTCHECK(regs);
+
 }
-#endif /*defined(FEATURE_INTERPRETIVE_EXECUTION)*/
 
 
 /*-------------------------------------------------------------------*/
 /* B238 RSCH  - Resume Subchannel                                [S] */
 /*-------------------------------------------------------------------*/
-void zz_resume_subchannel (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(resume_subchannel)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 DEVBLK *dev;                            /* -> device block           */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
     /* Program check if reg 1 bits 0-15 not X'0001' */
-    if ( (regs->gpr[1] >> 16) != 0x0001 )
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+    if ( regs->GR_LHH(1) != 0x0001 )
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
     /* Locate the device block for this subchannel */
-    dev = find_device_by_subchan (regs->gpr[1] & 0xFFFF);
+    dev = find_device_by_subchan (regs->GR_LHL(1));
 
     /* Condition code 3 if subchannel does not exist,
        is not valid, or is not enabled */
@@ -287,69 +292,66 @@ DEVBLK *dev;                            /* -> device block           */
     /* Perform resume subchannel and set condition code */
     regs->psw.cc = resume_subchan (regs, dev);
 
+    regs->siocount++;
 }
 
 
 /*-------------------------------------------------------------------*/
 /* B237 SAL   - Set Address Limit                                [S] */
 /*-------------------------------------------------------------------*/
-void zz_set_address_limit (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(set_address_limit)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
-    if(regs->gpr[1] & 0x8000FFFF)
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+    if(regs->GR_L(1) & 0x8000FFFF)
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
     else
-        sysblk.addrlimval = regs->gpr[1];
+        sysblk.addrlimval = regs->GR_L(1);
 }
 
 
 /*-------------------------------------------------------------------*/
 /* B23C SCHM  - Set Channel Monitor                              [S] */
 /*-------------------------------------------------------------------*/
-void zz_set_channel_monitor (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(set_channel_monitor)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
     /* Reserved bits in gpr1 must be zero */
-    if (regs->gpr[1] & CHM_GPR1_RESV)
-        program_interrupt (regs, PGM_SPECIFICATION_EXCEPTION);
+    if (regs->GR_L(1) & CHM_GPR1_RESV)
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
     /* Program check if M bit one and gpr2 address not on
-       a 32 byte boundary or highorder bit set */
-    if ((regs->gpr[1] & CHM_GPR1_M)
-     && (regs->gpr[2] & CHM_GPR2_RESV))
-        program_interrupt (regs, PGM_SPECIFICATION_EXCEPTION);
+       a 32 byte boundary or highorder bit set in ESA/390 mode */
+    if ((regs->GR_L(1) & CHM_GPR1_M)
+     && (regs->GR_L(2) & CHM_GPR2_RESV))
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
     /* Set the measurement block origin address */
-    if (regs->gpr[1] & CHM_GPR1_M)
+    if (regs->GR_L(1) & CHM_GPR1_M)
     {
-        sysblk.mbo = regs->gpr[2] & CHM_GPR2_MBO;
-        sysblk.mbk = (regs->gpr[1] & CHM_GPR1_MBK) >> 24;
+        sysblk.mbo = regs->GR(2);
+        sysblk.mbk = (regs->GR_L(1) & CHM_GPR1_MBK) >> 24;
         sysblk.mbm = 1;
     }
     else
         sysblk.mbm = 0;
 
-    sysblk.mbd = regs->gpr[1] & CHM_GPR1_D;
+    sysblk.mbd = regs->GR_L(1) & CHM_GPR1_D;
 
 }
 
@@ -357,19 +359,15 @@ U32     effective_addr2;                /* Effective address         */
 /*-------------------------------------------------------------------*/
 /* B233 SSCH  - Start Subchannel                                 [S] */
 /*-------------------------------------------------------------------*/
-void zz_start_subchannel (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(start_subchannel)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 DEVBLK *dev;                            /* -> device block           */
-U32     ccwaddr;                        /* CCW address for start I/O */
-U32     ioparm;                         /* I/O interruption parameter*/
 ORB     orb;                            /* Operation request block   */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
@@ -377,31 +375,33 @@ ORB     orb;                            /* Operation request block   */
     FW_CHECK(effective_addr2, regs);
 
     /* Fetch the operation request block */
-    vfetchc ( &orb, sizeof(ORB)-1, effective_addr2, b2, regs );
+    ARCH_DEP(vfetchc) ( &orb, sizeof(ORB)-1, effective_addr2, b2, regs );
 
     /* Program check if reserved bits are not zero */
-    if (orb.flag4 & ORB4_RESV
-        || orb.flag5 & ORB5_RESV
+    if (orb.flag5 & ORB5_RESV
         || orb.flag7 & ORB7_RESV
         || orb.ccwaddr[0] & 0x80)
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
+#if !defined(FEATURE_INCORRECT_LENGTH_INDICATION_SUPPRESSION)
     /* Program check if incorrect length suppression */
     if (orb.flag7 & ORB7_L)
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
+#endif /*!defined(FEATURE_INCORRECT_LENGTH_INDICATION_SUPPRESSION)*/
 
     /* Program check if reg 1 bits 0-15 not X'0001' */
-    if ( (regs->gpr[1] >> 16) != 0x0001 )
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+    if ( regs->GR_LHH(1) != 0x0001 )
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
     /* Locate the device block for this subchannel */
-    dev = find_device_by_subchan (regs->gpr[1] & 0xFFFF);
+    dev = find_device_by_subchan (regs->GR_LHL(1));
 
     /* Condition code 3 if subchannel does not exist,
-       is not valid, or is not enabled */
+       is not valid, is not enabled, or no path available */
     if (dev == NULL
         || (dev->pmcw.flag5 & PMCW5_V) == 0
-        || (dev->pmcw.flag5 & PMCW5_E) == 0)
+        || (dev->pmcw.flag5 & PMCW5_E) == 0
+        || (orb.lpm & dev->pmcw.pam) == 0)
     {
         regs->psw.cc = 3;
         return;
@@ -414,50 +414,43 @@ ORB     orb;                            /* Operation request block   */
     /* Clear the path not operational mask */
     dev->pmcw.pnom = 0;
 
-    /* Extract CCW address and I/O parameter */
-    ccwaddr = (orb.ccwaddr[0] << 24) | (orb.ccwaddr[1] << 16)
-                | (orb.ccwaddr[2] << 8) | orb.ccwaddr[3];
-    ioparm = (orb.intparm[0] << 24) | (orb.intparm[1] << 16)
-                | (orb.intparm[2] << 8) | orb.intparm[3];
+    /* Copy the logical path mask */
+    dev->pmcw.lpm = orb.lpm;
 
     /* Start the channel program and set the condition code */
-    WATCH("bef start");
-    regs->psw.cc =
-        start_io (dev, ioparm, orb.flag4, orb.flag5,
-                        orb.lpm, orb.flag7, ccwaddr);
+    regs->psw.cc = ARCH_DEP(startio) (dev, &orb);              /*@IWZ*/
 
-    WATCH("aft start");
-    /* Bump the I/O counter */
     regs->siocount++;
+
+    /* Set the last path used mask */
+    if (0 == regs->psw.cc) dev->pmcw.lpum = 0x80;
 }
 
 
 /*-------------------------------------------------------------------*/
 /* B23A STCPS - Store Channel Path Status                        [S] */
 /*-------------------------------------------------------------------*/
-void zz_store_channel_path_status (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(store_channel_path_status)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 BYTE    work[32];                       /* Work area                 */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
     /* Program check if operand not on 32 byte boundary */
     if ( effective_addr2 & 0x0000001F )
-        program_interrupt (regs, PGM_SPECIFICATION_EXCEPTION);
+        ARCH_DEP(program_interrupt) (regs, PGM_SPECIFICATION_EXCEPTION);
 
     /*INCOMPLETE, SET TO ALL ZEROS*/
     memset(work,0x00,32);
 
     /* Store channel path status word at operand address */
-    vstorec ( work, 32-1, effective_addr2, b2, regs );
+    ARCH_DEP(vstorec) ( work, 32-1, effective_addr2, b2, regs );
 
 }
 
@@ -465,27 +458,29 @@ BYTE    work[32];                       /* Work area                 */
 /*-------------------------------------------------------------------*/
 /* B239 STCRW - Store Channel Report Word                        [S] */
 /*-------------------------------------------------------------------*/
-void zz_store_channel_report_word (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(store_channel_report_word)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 U32     n;                              /* Integer work area         */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
     FW_CHECK(effective_addr2, regs);
 
+    /* Validate write access to operand before taking any
+       pending channel report word off the queue */
+    ARCH_DEP(validate_operand) (effective_addr2, b2, 0, ACCTYPE_WRITE, regs);
+
     /* Obtain any pending channel report */
     n = channel_report();
 
     /* Store channel report word at operand address */
-    vstore4 ( n, effective_addr2, b2, regs );
+    ARCH_DEP(vstore4) ( n, effective_addr2, b2, regs );
 
     /* Indicate if channel report or zeros were stored */
     regs->psw.cc = (n == 0) ? 1 : 0;
@@ -496,27 +491,25 @@ U32     n;                              /* Integer work area         */
 /*-------------------------------------------------------------------*/
 /* B234 STSCH - Store Subchannel                                 [S] */
 /*-------------------------------------------------------------------*/
-void zz_store_subchannel (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(store_subchannel)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 DEVBLK *dev;                            /* -> device block           */
 SCHIB   schib;                          /* Subchannel information blk*/
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
     /* Program check if reg 1 bits 0-15 not X'0001' */
-    if ( (regs->gpr[1] >> 16) != 0x0001 )
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+    if ( regs->GR_LHH(1) != 0x0001 )
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
     /* Locate the device block for this subchannel */
-    dev = find_device_by_subchan (regs->gpr[1] & 0xFFFF);
+    dev = find_device_by_subchan (regs->GR_LHL(1));
 
     /* Set condition code 3 if subchannel does not exist */
     if (dev == NULL)
@@ -537,7 +530,7 @@ SCHIB   schib;                          /* Subchannel information blk*/
     memset (schib.moddep, 0, sizeof(schib.moddep));
 
     /* Store the subchannel information block */
-    vstorec ( &schib, sizeof(SCHIB)-1, effective_addr2,
+    ARCH_DEP(vstorec) ( &schib, sizeof(SCHIB)-1, effective_addr2,
                 b2, regs );
 
     /* Set condition code 0 */
@@ -549,19 +542,18 @@ SCHIB   schib;                          /* Subchannel information blk*/
 /*-------------------------------------------------------------------*/
 /* B236 TPI   - Test Pending Interruption                        [S] */
 /*-------------------------------------------------------------------*/
-void zz_test_pending_interruption (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(test_pending_interruption)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 PSA    *psa;                            /* -> Prefixed storage area  */
 U64     dreg;                           /* Double register work area */
 U32     ioid;                           /* I/O interruption address  */
 U32     ioparm;                         /* I/O interruption parameter*/
+U32     iointid;                        /* I/O interruption ident    */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
@@ -570,63 +562,66 @@ U32     ioparm;                         /* I/O interruption parameter*/
 
     /* validate operand before taking any action */
     if ( effective_addr2 != 0 )
-        validate_operand (effective_addr2, b2, 8-1, ACCTYPE_WRITE, regs);
+        ARCH_DEP(validate_operand) (effective_addr2, b2, 8-1,
+                                                  ACCTYPE_WRITE, regs);
 
     /* Perform serialization and checkpoint-synchronization */
     PERFORM_SERIALIZATION (regs);
     PERFORM_CHKPT_SYNC (regs);
 
-    /* Obtain the interrupt lock */
-    obtain_lock (&sysblk.intlock);
-
-    /* Test and clear pending interrupt, set condition code */
-    regs->psw.cc =
-        present_io_interrupt (regs, &ioid, &ioparm, NULL);
-
-    /* Release the interrupt lock */
-    release_lock (&sysblk.intlock);
-
-    /* Store the SSID word and I/O parameter if an interrupt was pending */
-    if (regs->psw.cc)
+    if( IS_IC_IOPENDING )
     {
-	if ( effective_addr2 == 0 )
-	{
-            /* If operand address is zero, store in PSA */
-            psa = (PSA*)(sysblk.mainstor + regs->pxr);
-            psa->ioid[0] = ioid >> 24;
-            psa->ioid[1] = (ioid & 0xFF0000) >> 16;
-            psa->ioid[2] = (ioid & 0xFF00) >> 8;
-            psa->ioid[3] = ioid & 0xFF;
-            psa->ioparm[0] = ioparm >> 24;
-            psa->ioparm[1] = (ioparm & 0xFF0000) >> 16;
-            psa->ioparm[2] = (ioparm & 0xFF00) >> 8;
-            psa->ioparm[3] = ioparm & 0xFF;
-	}
-	else
-	{
-            /* Otherwise store at operand location */
-            dreg = ((U64)ioid << 32) | ioparm;
-            vstore8 ( dreg, effective_addr2, b2, regs );
-	}
-    }
+        /* Obtain the interrupt lock */
+        obtain_lock (&sysblk.intlock);
 
+        /* Test and clear pending interrupt, set condition code */
+        regs->psw.cc =
+            ARCH_DEP(present_io_interrupt) (regs, &ioid, &ioparm,
+                                                       &iointid, NULL);
+
+        /* Release the interrupt lock */
+        release_lock (&sysblk.intlock);
+
+        /* Store the SSID word and I/O parameter if an interrupt was pending */
+        if (regs->psw.cc)
+        {
+            if ( effective_addr2 == 0 )
+            {
+                /* If operand address is zero, store in PSA */
+                psa = (void*)(sysblk.mainstor + regs->PX);
+                STORE_FW(psa->ioid,ioid);
+                STORE_FW(psa->ioparm,ioparm);
+#if defined(FEATURE_ESAME)
+                STORE_FW(psa->iointid,iointid);
+#endif /*defined(FEATURE_ESAME)*/
+                STORAGE_KEY(regs->PX) |= (STORKEY_REF|STORKEY_CHANGE);
+            }
+            else
+            {
+                /* Otherwise store at operand location */
+                dreg = ((U64)ioid << 32) | ioparm;
+                ARCH_DEP(vstore8) ( dreg, effective_addr2, b2, regs );
+            }
+        }
+    }
+    else
+        regs->psw.cc = 0;
+    
 }
 
 
 /*-------------------------------------------------------------------*/
 /* B235 TSCH  - Test Subchannel                                  [S] */
 /*-------------------------------------------------------------------*/
-void zz_test_subchannel (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(test_subchannel)
 {
 int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 DEVBLK *dev;                            /* -> device block           */
 IRB     irb;                            /* Interruption response blk */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_XA(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
@@ -634,11 +629,11 @@ IRB     irb;                            /* Interruption response blk */
     FW_CHECK(effective_addr2, regs);
 
     /* Program check if reg 1 bits 0-15 not X'0001' */
-    if ( (regs->gpr[1] >> 16) != 0x0001 )
-        program_interrupt (regs, PGM_OPERAND_EXCEPTION);
+    if ( regs->GR_LHH(1) != 0x0001 )
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
 
     /* Locate the device block for this subchannel */
-    dev = find_device_by_subchan (regs->gpr[1] & 0xFFFF);
+    dev = find_device_by_subchan (regs->GR_LHL(1));
 
     /* Condition code 3 if subchannel does not exist,
        is not valid, or is not enabled */
@@ -651,7 +646,7 @@ IRB     irb;                            /* Interruption response blk */
     }
 
     /* validate operand before taking any action */
-    validate_operand (effective_addr2, b2, sizeof(IRB)-1,
+    ARCH_DEP(validate_operand) (effective_addr2, b2, sizeof(IRB)-1,
                                         ACCTYPE_WRITE, regs);
 
     /* Perform serialization and checkpoint-synchronization */
@@ -662,39 +657,74 @@ IRB     irb;                            /* Interruption response blk */
     regs->psw.cc = test_subchan (regs, dev, &irb);
 
     /* Store the interruption response block */
-    vstorec ( &irb, sizeof(IRB)-1, effective_addr2, b2, regs );
+    ARCH_DEP(vstorec) ( &irb, sizeof(IRB)-1, effective_addr2, b2, regs );
 
 }
 
-#endif /*defined(FEATURE_CHANNEL_SUBSYSTEM)*/
-
-
-#if defined(FEATURE_S370_CHANNEL) || defined(FEATURE_INTERPRETIVE_EXECUTION)
 
+#if defined(FEATURE_CANCEL_IO_FACILITY)
 /*-------------------------------------------------------------------*/
-/* 9C00 SIO   - Start I/O                                        [S] */
-/* 9C01 SIOF  - Start I/O Fast Release                           [S] */
+/* B276 XSCH  - Cancel Subchannel                                [S] */
 /*-------------------------------------------------------------------*/
-void zz_s370_startio (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(cancel_subchannel)
 {
-int     b2;                             /* Effective addr base       */
-U32     effective_addr2;                /* Effective address         */
-#if !defined(FEATURE_INTERPRETIVE_EXECUTION)
-PSA    *psa;                            /* -> prefixed storage area  */
-DEVBLK *dev;                            /* -> device block for SIO   */
-U32     ccwaddr;                        /* CCW address for start I/O */
-BYTE    ccwkey;                         /* Bits 0-3=key, 4=7=zeroes  */
-#endif /*!defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+int     b2;                             /* Base of effective addr    */
+VADR    effective_addr2;                /* Effective address         */
+DEVBLK *dev;                            /* -> device block           */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_370(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
-#if !defined(FEATURE_INTERPRETIVE_EXECUTION)
+    /* Program check if reg 1 bits 0-15 not X'0001' */
+    if ( regs->GR_LHH(1) != 0x0001 )
+        ARCH_DEP(program_interrupt) (regs, PGM_OPERAND_EXCEPTION);
+
+    /* Locate the device block for this subchannel */
+    dev = find_device_by_subchan (regs->GR_LHL(1));
+
+    /* Condition code 3 if subchannel does not exist,
+       is not valid, or is not enabled */
+    if (dev == NULL
+        || (dev->pmcw.flag5 & PMCW5_V) == 0
+        || (dev->pmcw.flag5 & PMCW5_E) == 0)
+    {
+        regs->psw.cc = 3;
+        return;
+    }
+
+    /* Perform halt subchannel and set condition code */
+    regs->psw.cc = cancel_subchan (regs, dev);
+
+}
+#endif /*defined(FEATURE_CANCEL_IO_FACILITY)*/
+#endif /*defined(FEATURE_CHANNEL_SUBSYSTEM)*/
+
+
+#if defined(FEATURE_S370_CHANNEL)
+
+/*-------------------------------------------------------------------*/
+/* 9C00 SIO   - Start I/O                                        [S] */
+/* 9C01 SIOF  - Start I/O Fast Release                           [S] */
+/* 9C02 RIO   - Resume I/O   -   ZZ INCOMPLETE                   [S] */
+/*-------------------------------------------------------------------*/
+DEF_INST(start_io)
+{
+int     b2;                             /* Effective addr base       */
+VADR    effective_addr2;                /* Effective address         */
+PSA    *psa;                            /* -> prefixed storage area  */
+DEVBLK *dev;                            /* -> device block for SIO   */
+ORB     orb;                            /* Operation request blk @IZW*/
+VADR    ccwaddr;                        /* CCW address for start I/O */
+BYTE    ccwkey;                         /* Bits 0-3=key, 4=7=zeroes  */
+
+    S(inst, execflag, regs, b2, effective_addr2);
+
+    PRIV_CHECK(regs);
+
+    SIE_INTERCEPT(regs);
 
     /* Locate the device block */
     dev = find_device_by_devnum (effective_addr2);
@@ -707,20 +737,21 @@ BYTE    ccwkey;                         /* Bits 0-3=key, 4=7=zeroes  */
     }
 
     /* Fetch key and CCW address from the CAW at PSA+X'48' */
-    psa = (PSA*)(sysblk.mainstor + regs->pxr);
+    psa = (PSA*)(sysblk.mainstor + regs->PX);
     ccwkey = psa->caw[0] & 0xF0;
     ccwaddr = (psa->caw[1] << 16) | (psa->caw[2] << 8)
                     | psa->caw[3];
 
+    /* Build the I/O operation request block */                /*@IZW*/
+    memset (&orb, 0, sizeof(ORB));                             /*@IZW*/
+    orb.flag4 = ccwkey & ORB4_KEY;                             /*@IZW*/
+    STORE_FW(orb.ccwaddr,ccwaddr);                             /*@IZW*/
+
     /* Start the channel program and set the condition code */
-    regs->psw.cc =
-        start_io (dev, 0, ccwkey, 0, 0, 0, ccwaddr);
+    regs->psw.cc = ARCH_DEP(startio) (dev, &orb);              /*@IZW*/
 
-    /* Bump the I/O counter */
     regs->siocount++;
 
-#endif /*!defined(FEATURE_INTERPRETIVE_EXECUTION)*/
-
 }
 
 
@@ -728,24 +759,18 @@ BYTE    ccwkey;                         /* Bits 0-3=key, 4=7=zeroes  */
 /* 9D00 TIO   - Test I/O                                         [S] */
 /* 9D01 CLRIO - Clear I/O                                        [S] */
 /*-------------------------------------------------------------------*/
-void zz_s370_testio (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(test_io)
 {
 int     b2;                             /* Base of effective addr    */
-U32     effective_addr2;                /* Effective address         */
-#if !defined(FEATURE_INTERPRETIVE_EXECUTION)
+VADR    effective_addr2;                /* Effective address         */
 DEVBLK *dev;                            /* -> device block for SIO   */
-#endif /*!defined(FEATURE_INTERPRETIVE_EXECUTION)*/
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_370(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
-#if !defined(FEATURE_INTERPRETIVE_EXECUTION)
-
     /* Locate the device block */
     dev = find_device_by_devnum (effective_addr2);
 
@@ -757,9 +782,7 @@ DEVBLK *dev;                            /* -> device block for SIO   */
     }
 
     /* Test the device and set the condition code */
-    regs->psw.cc = test_io (regs, dev, inst[1]);
-
-#endif /*!defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+    regs->psw.cc = testio (regs, dev, inst[1]);
 
 }
 
@@ -768,24 +791,18 @@ DEVBLK *dev;                            /* -> device block for SIO   */
 /* 9E00 HIO   - Halt I/O                                         [S] */
 /* 9E01 HDV   - Halt Device                                      [S] */
 /*-------------------------------------------------------------------*/
-void zz_s370_haltio (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(halt_io)
 {
 int     b2;                             /* Base of effective addr    */
-U32     effective_addr2;                /* Effective address         */
-#if !defined(FEATURE_INTERPRETIVE_EXECUTION)
+VADR    effective_addr2;                /* Effective address         */
 DEVBLK *dev;                            /* -> device block for SIO   */
-#endif /*!defined(FEATURE_INTERPRETIVE_EXECUTION)*/
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_370(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
-#if !defined(FEATURE_INTERPRETIVE_EXECUTION)
-
     /* Locate the device block */
     dev = find_device_by_devnum (effective_addr2);
 
@@ -797,9 +814,7 @@ DEVBLK *dev;                            /* -> device block for SIO   */
     }
 
     /* Test the device and set the condition code */
-    regs->psw.cc = halt_io (regs, dev, inst[1]);
-
-#endif /*!defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+    regs->psw.cc = haltio (regs, dev, inst[1]);
 
 }
 
@@ -807,37 +822,38 @@ DEVBLK *dev;                            /* -> device block for SIO   */
 /*-------------------------------------------------------------------*/
 /* 9F00 TCH   - Test Channel                                     [S] */
 /*-------------------------------------------------------------------*/
-void zz_s370_test_channel (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(test_channel)
 {
 int     b2;                             /* Base of effective addr    */
-U32     effective_addr2;                /* Effective address         */
-#if defined(FEATURE_INTERPRETIVE_EXECUTION)
+VADR    effective_addr2;                /* Effective address         */
+#if defined(_FEATURE_SIE)
 BYTE    channelid;
 U16     tch_ctl;
-#endif /*defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+#endif /*defined(_FEATURE_SIE)*/
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_370(regs);
-
     PRIV_CHECK(regs);
 
-#if !defined(FEATURE_INTERPRETIVE_EXECUTION)
-
-    /* Test for pending interrupt and set condition code */
-    regs->psw.cc = test_channel (regs, effective_addr2 & 0xFF00);
-
-#else /*defined(FEATURE_INTERPRETIVE_EXECUTION)*/
-
-    channelid = (effective_addr2 >> 8) & 0xFF;
-    tch_ctl = regs->siebk->tch_ctl[0] << 8 | regs->siebk->tch_ctl[1];
-    if((channelid > 15)
-     || ((0x8000 >> channelid) & tch_ctl))
-        longjmp(regs->progjmp, SIE_INTERCEPT_INST);
+#if defined(_FEATURE_SIE)
+    if(!regs->sie_state)
+    {
+#endif /*defined(_FEATURE_SIE)*/
+        /* Test for pending interrupt and set condition code */
+        regs->psw.cc = testch (regs, effective_addr2 & 0xFF00);
+#if defined(_FEATURE_SIE)
+    }
     else
-        regs->psw.cc = 0;
-
-#endif /*defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+    {
+        channelid = (effective_addr2 >> 8) & 0xFF;
+        FETCH_HW(tch_ctl,((SIEBK*)(regs->siebk))->tch_ctl);
+        if((channelid > 15)
+         || ((0x8000 >> channelid) & tch_ctl))
+            longjmp(regs->progjmp, SIE_INTERCEPT_INST);
+        else
+            regs->psw.cc = 0;
+    }
+#endif /*defined(_FEATURE_SIE)*/
 
 }
 
@@ -845,47 +861,36 @@ U16     tch_ctl;
 /*-------------------------------------------------------------------*/
 /* B203 STIDC - Store Channel ID                                 [S] */
 /*-------------------------------------------------------------------*/
-void zz_s370_store_channelid (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(store_channel_id)
 {
 int     b2;                             /* Base of effective addr    */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    SIE_MODE_370(regs);
-
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
-#if !defined(FEATURE_INTERPRETIVE_EXECUTION)
-
     /* Store Channel ID and set condition code */
     regs->psw.cc =
-        store_channel_id (regs, effective_addr2 & 0xFF00);
-
-#endif /*!defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+        stchan_id (regs, effective_addr2 & 0xFF00);
 
 }
 
-#endif /*defined(FEATURE_S370_CHANNEL)*/
 
-
-#if defined(FEATURE_INTERPRETIVE_EXECUTION)
+#if defined(_FEATURE_SIE)
 /*-------------------------------------------------------------------*/
 /* B200 CONCS - Connect Channel Set                              [S] */
 /*-------------------------------------------------------------------*/
-void zz_s370_connect_channel_set (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(connect_channel_set)
 {
 int     b2;                             /* Base of effective addr    */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    if(!regs->sie_state)
-        program_interrupt(regs, PGM_OPERATION_EXCEPTION);
-
-    SIE_MODE_370(regs);
+    SIE_ONLY_INSTRUCTION(regs);
 
     PRIV_CHECK(regs);
 
@@ -897,21 +902,31 @@ U32     effective_addr2;                /* Effective address         */
 /*-------------------------------------------------------------------*/
 /* B201 DISCS - Disconnect Channel Set                           [S] */
 /*-------------------------------------------------------------------*/
-void zz_s370_disconnect_channel_set (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(disconnect_channel_set)
 {
 int     b2;                             /* Base of effective addr    */
-U32     effective_addr2;                /* Effective address         */
+VADR    effective_addr2;                /* Effective address         */
 
     S(inst, execflag, regs, b2, effective_addr2);
 
-    if(!regs->sie_state)
-        program_interrupt(regs, PGM_OPERATION_EXCEPTION);
-
-    SIE_MODE_370(regs);
+    SIE_ONLY_INSTRUCTION(regs);
 
     PRIV_CHECK(regs);
 
     SIE_INTERCEPT(regs);
 
 }
-#endif /*defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+#endif /*defined(_FEATURE_SIE)*/
+#endif /*defined(FEATURE_S370_CHANNEL)*/
+
+
+#if !defined(_GEN_ARCH)
+
+#define  _GEN_ARCH 390
+#include "io.c"
+
+#undef   _GEN_ARCH
+#define  _GEN_ARCH 370
+#include "io.c"
+
+#endif /*!defined(_GEN_ARCH)*/
