@@ -1,4 +1,4 @@
-/* PRINTER.C    (c) Copyright Roger Bowler, 1999                     */
+/* PRINTER.C    (c) Copyright Roger Bowler, 1999-2001                */
 /*              ESA/390 Line Printer Device Handler                  */
 
 /*-------------------------------------------------------------------*/
@@ -149,7 +149,7 @@ BYTE            c;                      /* Print character           */
     if (dev->fd < 0 && !IS_CCW_SENSE(code))
     {
         rc = open (dev->filename,
-                    O_WRONLY | O_CREAT | O_TRUNC | O_SYNC,
+                    O_WRONLY | O_CREAT | O_TRUNC /* | O_SYNC */,
                     S_IRUSR | S_IWUSR | S_IRGRP);
         if (rc < 0)
         {
