@@ -1,9 +1,9 @@
 /* XSTORE.C   Expanded storage related instructions - Jan Jaeger     */
 
-/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2000      */
+/* Interpretive Execution - (c) Copyright Jan Jaeger, 1999-2001      */
+/* z/Architecture support - (c) Copyright Jan Jaeger, 1999-2001      */
 
 /* MVPG moved from cpu.c to xstore.c   05/07/00 Jan Jaeger */
-/* Lock Page added                     29/07/00 Jan Jaeger */
 
 #include "hercules.h"
 
@@ -15,20 +15,20 @@
 /*-------------------------------------------------------------------*/
 /* B22E PGIN  - Page in from expanded storage                  [RRE] */
 /*-------------------------------------------------------------------*/
-void zz_page_in (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(page_in)
 {
 int     r1, r2;                         /* Values of R fields        */
-U32     maddr;                          /* Main storage address      */
+RADR    maddr;                          /* Main storage address      */
 U32     xaddr;                          /* Expanded storage address  */
 
     RRE(inst, execflag, regs, r1, r2);
 
     PRIV_CHECK(regs);
 
-#if defined(FEATURE_INTERPRETIVE_EXECUTION)
+#if defined(_FEATURE_SIE)
     if(regs->sie_state && (regs->siebk->ic[3] & SIE_IC3_PGX))
         longjmp(regs->progjmp, SIE_INTERCEPT_INST);
-#endif /*defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+#endif /*defined(_FEATURE_SIE)*/
 
 #if defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE)
     /* Cannot perform xstore page movement in XC mode */
@@ -37,9 +37,9 @@ U32     xaddr;                          /* Expanded storage address  */
 #endif /*defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE)*/
 
     /* expanded storage block number */
-    xaddr = regs->gpr[r2];
+    xaddr = regs->GR_L(r2);
 
-#if defined(FEATURE_INTERPRETIVE_EXECUTION)
+#if defined(_FEATURE_SIE)
     if(regs->sie_state)
     {
         xaddr += regs->sie_xso;
@@ -49,7 +49,7 @@ U32     xaddr;                          /* Expanded storage address  */
             return;
         }
     }
-#endif /*defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+#endif /*defined(_FEATURE_SIE)*/
 
     /* If the expanded storage block is not configured then
        terminate with cc3 */
@@ -63,8 +63,8 @@ U32     xaddr;                          /* Expanded storage address  */
     xaddr <<= XSTORE_PAGESHIFT;
 
     /* Obtain abs address, verify access and set ref/change bits */
-    maddr = logical_to_abs (regs->gpr[r1] & ADDRESS_MAXWRAP(regs),
-         USE_REAL_ADDR, regs, ACCTYPE_WRITE, regs->psw.pkey);
+    maddr = LOGICAL_TO_ABS (regs->GR(r1) & ADDRESS_MAXWRAP(regs),
+         USE_REAL_ADDR, regs, ACCTYPE_WRITE, 0);
     maddr &= XSTORE_PAGEMASK;
 
     /* Copy data from expanded to main */
@@ -74,30 +74,28 @@ U32     xaddr;                          /* Expanded storage address  */
     /* cc0 means pgin ok */
     regs->psw.cc = 0;
 
-#ifdef IBUF
-    FRAG_INVALIDATE(maddr, XSTORE_PAGESIZE);
-#endif
-
 } 
+#endif /*defined(FEATURE_EXPANDED_STORAGE)*/
 
 
+#if defined(FEATURE_EXPANDED_STORAGE)
 /*-------------------------------------------------------------------*/
 /* B22F PGOUT - Page out to expanded storage                   [RRE] */
 /*-------------------------------------------------------------------*/
-void zz_page_out (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(page_out)
 {
 int     r1, r2;                         /* Values of R fields        */
-U32     maddr;                          /* Main storage address      */
+RADR    maddr;                          /* Main storage address      */
 U32     xaddr;                          /* Expanded storage address  */
 
     RRE(inst, execflag, regs, r1, r2);
 
     PRIV_CHECK(regs);
 
-#if defined(FEATURE_INTERPRETIVE_EXECUTION)
+#if defined(_FEATURE_SIE)
     if(regs->sie_state && (regs->siebk->ic[3] & SIE_IC3_PGX))
         longjmp(regs->progjmp, SIE_INTERCEPT_INST);
-#endif /*defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+#endif /*defined(_FEATURE_SIE)*/
 
 #if defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE)
     /* Cannot perform xstore page movement in XC mode */
@@ -106,9 +104,9 @@ U32     xaddr;                          /* Expanded storage address  */
 #endif /*defined(FEATURE_MULTIPLE_CONTROLLED_DATA_SPACE)*/
 
     /* expanded storage block number */
-    xaddr = regs->gpr[r2];
+    xaddr = regs->GR_L(r2);
 
-#if defined(FEATURE_INTERPRETIVE_EXECUTION)
+#if defined(_FEATURE_SIE)
     if(regs->sie_state)
     {
         xaddr += regs->sie_xso;
@@ -118,7 +116,7 @@ U32     xaddr;                          /* Expanded storage address  */
             return;
         }
     }
-#endif /*defined(FEATURE_INTERPRETIVE_EXECUTION)*/
+#endif /*defined(_FEATURE_SIE)*/
 
     /* If the expanded storage block is not configured then
        terminate with cc3 */
@@ -132,7 +130,7 @@ U32     xaddr;                          /* Expanded storage address  */
     xaddr <<= XSTORE_PAGESHIFT;
 
     /* Obtain abs address, verify access and set ref/change bits */
-    maddr = logical_to_abs (regs->gpr[r1] & ADDRESS_MAXWRAP(regs),
+    maddr = LOGICAL_TO_ABS (regs->GR(r1) & ADDRESS_MAXWRAP(regs),
          USE_REAL_ADDR, regs, ACCTYPE_READ, regs->psw.pkey);
     maddr &= XSTORE_PAGEMASK;
 
@@ -147,57 +145,11 @@ U32     xaddr;                          /* Expanded storage address  */
 #endif /*defined(FEATURE_EXPANDED_STORAGE)*/
 
 
-/*-------------------------------------------------------------------*/
-/* B262 LKPG  - Lock Page                                      [RRE] */
-/*-------------------------------------------------------------------*/
-void zz_lock_page (BYTE inst[], int execflag, REGS *regs)
-{
-int     r1, r2;                         /* Values of R fields        */
-U32     n2;                             /* effective addr of r2      */
-U32     raddr;                          /* Real address              */
-int     private;                        /* 1=Private address space   */
-int     protect;                        /* 1=ALE or page protection  */
-int     stid;                           /* Segment table indication  */
-U16     xcode;                          /* Exception code            */
-
-    RRE(inst, execflag, regs, r1, r2);
-
-    PRIV_CHECK(regs);
-
-    if(REAL_MODE(&(regs->psw)))
-        program_interrupt (regs, PGM_SPECIAL_OPERATION_EXCEPTION);
-
-    n2 = regs->gpr[r2] & ADDRESS_MAXWRAP(regs);
-
-    if(regs->gpr[0] & LKPG_GPR0_RESV)
-        program_interrupt (regs, PGM_SPECIFICATION_EXCEPTION);
-
-    n2 = regs->gpr[r2] & ADDRESS_MAXWRAP(regs);
-
-    /* Access to PTE must be serialized */
-    OBTAIN_MAINLOCK(regs);
-
-    /* Return condition code 3 if translation exception */
-    if (translate_addr (n2, r2, regs, (regs->gpr[0] & LKPG_GPR0_LOCKBIT)
-                ? ACCTYPE_LOCKPAGE : ACCTYPE_UNLKPAGE,
-                &raddr, &xcode, &private, &protect, &stid, NULL, NULL))
-        regs->psw.cc = 3;
-
-    RELEASE_MAINLOCK(regs);
-
-    if((regs->gpr[0] & LKPG_GPR0_LOCKBIT) && regs->psw.cc == 0)
-        regs->gpr[r1] = raddr;
-
-}
-
-
-#if defined(FEATURE_MOVE_PAGE_FACILITY_2)
-
-
+#if defined(FEATURE_MOVE_PAGE_FACILITY_2) && defined(FEATURE_EXPANDED_STORAGE)
 /*-------------------------------------------------------------------*/
 /* B259 IESBE - Invalidate Expanded Storage Blk Entry          [RRE] */
 /*-------------------------------------------------------------------*/
-void zz_invalidate_expanded_storage_block_entry (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(invalidate_expanded_storage_block_entry)
 {
 int     r1, r2;                         /* Values of R fields        */
 
@@ -212,33 +164,19 @@ int     r1, r2;                         /* Values of R fields        */
     OBTAIN_MAINLOCK(regs);
 
     /* Invalidate page table entry */
-    invalidate_pte (inst[1], r1, r2, regs);
+    ARCH_DEP(invalidate_pte) (inst[1], r1, r2, regs);
 
-    /* Release mainstore interlock */
-    RELEASE_MAINLOCK(regs);
+    /* Mainlock now released by `invalidate_pte' */
+//  RELEASE_MAINLOCK(regs);
 
     /* Perform serialization after operation */
     PERFORM_SERIALIZATION (regs);
 
 }
+#endif /*defined(FEATURE_EXPANDED_STORAGE)*/
 
 
-/*-------------------------------------------------------------------*/
-/* Subroutine to locate page in expanded storage                     */
-/*                                                                   */
-/* Input:                                                            */
-/*      pteadr  Real address of page table entry                     */
-/*      regs    Pointer to the CPU register context                  */
-/* Output:                                                           */
-/*      xpblkn  Expanded storage block number if valid               */
-/* Return value:                                                     */
-/*      Returns 1 if page is valid in expanded storage, else 0.      */
-/*-------------------------------------------------------------------*/
-static inline int locate_expanded_page (U32 pteadr, U32 *xpblkn, REGS *regs)
-{
-    return 0;
-} /* end function locate_expanded_page */
-
+#if defined(FEATURE_MOVE_PAGE_FACILITY_2)
 /*-------------------------------------------------------------------*/
 /* Move Page (Facility 2)                                            */
 /*                                                                   */
@@ -255,65 +193,62 @@ static inline int locate_expanded_page (U32 pteadr, U32 *xpblkn, REGS *regs)
 /*-------------------------------------------------------------------*/
 /* B254 MVPG  - Move Page                                      [RRE] */
 /*-------------------------------------------------------------------*/
-void zz_move_page (BYTE inst[], int execflag, REGS *regs)
+DEF_INST(move_page)
 {
 int     r1, r2;                         /* Register values           */
-int     rc;                             /* Return code               */
-U32     vaddr1, vaddr2;                 /* Virtual addresses         */
-U32     raddr1, raddr2;                 /* Real addresses            */
-U32     aaddr1 = 0, aaddr2 = 0;         /* Absolute addresses        */
-int     xpvalid1 = 0, xpvalid2 = 0;     /* 1=Operand in expanded stg */
-U32     xpblk1, xpblk2;                 /* Expanded storage block#   */
-int     priv = 0;                       /* 1=Private address space   */
+int     rc = 0;                         /* Return code               */
+int     cc = 0;				/* Condition code            */
+VADR    vaddr1, vaddr2;                 /* Virtual addresses         */
+RADR    raddr1, raddr2;                 /* Real addresses            */
+RADR    aaddr1 = 0, aaddr2 = 0;         /* Absolute addresses        */
+int     priv;                           /* 1=Private address space   */
 int     prot = 0;                       /* 1=Protected page          */
 int     stid;                           /* Segment table indication  */
-U32     xaddr;                          /* Address causing exception */
 U16     xcode;                          /* Exception code            */
 BYTE    akey;                           /* Access key in register 0  */
 BYTE    akey1, akey2;                   /* Access keys for operands  */
-BYTE    xpkey1, xpkey2;                 /* Expanded storage keys     */
+#if defined(FEATURE_EXPANDED_STORAGE)
+int     xpvalid1 = 0, xpvalid2 = 0;     /* 1=Operand in expanded stg */
+CREG    pte1 = 0, pte2 = 0;             /* Page table entry          */
+U32     xpblk1 = 0, xpblk2 = 0;         /* Expanded storage block#   */
+BYTE    xpkey1 = 0, xpkey2 = 0;         /* Expanded storage keys     */
+#endif /*defined(FEATURE_EXPANDED_STORAGE)*/
 
     RRE(inst, execflag, regs, r1, r2);
 
     /* Specification exception if register 0 bits 16-19 are
        not all zero, or if bits 20 and 21 are both ones */
-    if ((regs->gpr[0] & 0x0000F000) != 0
-        || (regs->gpr[0] & 0x00000C00) == 0x00000C00)
-    {
-        program_interrupt (regs, PGM_SPECIFICATION_EXCEPTION);
-        return 3;
-    }
+    if ((regs->GR_L(0) & 0x0000F000) != 0
+        || (regs->GR_L(0) & 0x00000C00) == 0x00000C00)
+        ARCH_DEP(program_interrupt) (regs, PGM_SPECIFICATION_EXCEPTION);
 
     /* Use PSW key as access key for both operands */
     akey1 = akey2 = regs->psw.pkey;
 
     /* If register 0 bit 20 or 21 is one, get access key from R0 */
-    if (regs->gpr[0] & 0x00000C00)
+    if (regs->GR_L(0) & 0x00000C00)
     {
         /* Extract the access key from register 0 bits 24-27 */
-        akey = regs->gpr[0] & 0x000000F0;
+        akey = regs->GR_L(0) & 0x000000F0;
 
         /* Priviliged operation exception if in problem state, and
            the specified key is not permitted by the PSW key mask */
         if ( regs->psw.prob
-            && ((regs->cr[3] << (akey >> 4)) & 0x80000000) == 0 )
-        {
-            program_interrupt (regs, PGM_PRIVILEGED_OPERATION_EXCEPTION);
-            return 3;
-        }
+            && ((regs->CR(3) << (akey >> 4)) & 0x80000000) == 0 )
+            ARCH_DEP(program_interrupt) (regs, PGM_PRIVILEGED_OPERATION_EXCEPTION);
 
         /* If register 0 bit 20 is one, use R0 key for operand 1 */
-        if (regs->gpr[0] & 0x00000800)
+        if (regs->GR_L(0) & 0x00000800)
             akey1 = akey;
 
         /* If register 0 bit 21 is one, use R0 key for operand 2 */
-        if (regs->gpr[0] & 0x00000400)
+        if (regs->GR_L(0) & 0x00000400)
             akey2 = akey;
     }
 
     /* Determine the logical addresses of each operand */
-    vaddr1 = regs->gpr[r1] & ADDRESS_MAXWRAP(regs);
-    vaddr2 = regs->gpr[r2] & ADDRESS_MAXWRAP(regs);
+    vaddr1 = regs->GR(r1) & ADDRESS_MAXWRAP(regs);
+    vaddr2 = regs->GR(r2) & ADDRESS_MAXWRAP(regs);
 
     /* Isolate the page addresses of each operand */
     vaddr1 &= XSTORE_PAGEMASK;
@@ -323,53 +258,166 @@ BYTE    xpkey1, xpkey2;                 /* Expanded storage keys     */
     if (!REAL_MODE(&regs->psw))
     {
         /* Translate the second operand address to a real address */
-        rc = translate_addr (vaddr2, r2, regs, ACCTYPE_READ, &raddr2,
-                        &xcode, &priv, &prot, &stid, &xpblk2, &xpkey2);
+        rc = ARCH_DEP(translate_addr) (vaddr2, r2, regs, ACCTYPE_READ, &raddr2,
+                        &xcode, &priv, &prot, &stid);
 
-        /* If page is invalid in real storage but valid in expanded
-           storage then xpblk2 now contains expanded storage block# */
-        if (rc == 5) xpvalid2 = 1;
+	raddr2 = APPLY_PREFIXING (raddr2, regs->PX);
+
+        if (raddr2 < regs->mainsize)
+            SIE_TRANSLATE(&raddr2,ACCTYPE_READ,regs);
+
+#if defined(FEATURE_EXPANDED_STORAGE)
+        if(rc == 2)
+        {
+	    FETCH_W(pte2,sysblk.mainstor + raddr2);
+            /* If page is invalid in real storage but valid in expanded
+               storage then xpblk2 now contains expanded storage block# */
+            if(pte2 & PAGETAB_ESVALID)
+            {
+                xpblk2 = pte2 & ZPGETAB_PFRA >> 12;
+#if defined(_FEATURE_SIE)
+                if(regs->sie_state)
+                {
+                    /* Add expanded storage origin for this guest */
+                    xpblk2 += regs->sie_xso;
+                    /* If the block lies beyond this guests limit
+                       then we must terminate the instruction */
+                    if(xpblk2 >= regs->sie_xsl)
+                    {
+                        cc = 2;
+                        goto mvpg_progck;
+                    }
+                }
+#endif /*defined(_FEATURE_SIE)*/
+
+                /* If the expanded storage block does not exist
+                   then terminate the instruction */
+                if (xpblk2 >= sysblk.xpndsize)
+                {
+                    cc = 2;
+                    goto mvpg_progck;
+                }
+
+                rc = 0;
+                xpvalid2 = 1;
+                xpkey2 = sysblk.mainstor[raddr2 +
+#if defined(FEATURE_ESAME)
+                                                  2048
+#else /*!defined(FEATURE_ESAME)*/
+                /* For ESA/390 mode, the XPTE lies directly beyond 
+                   the PTE, and each entry is 12 bytes long, we must
+                   therefor add 1024 + 8 times the page index */
+	                             1024 + ((vaddr1 & 0x000FF000) >> 9)
+#endif /*!defined(FEATURE_ESAME)*/
+                                                      ];
+/*DEBUG*/ logmsg("MVPG pte2 = " F_CREG ", xkey2 = %2.2X\n",pte2,xpkey2);
+            }
+            else
+            {
+                cc = 2;
+                goto mvpg_progck;
+            }
+        }
+#endif /*defined(FEATURE_EXPANDED_STORAGE)*/
 
         /* Program check if second operand is not valid
            in either main storage or expanded storage */
-        if (rc != 0 && xpvalid2 == 0)
+        if (rc)
         {
-            xaddr = vaddr2;
+	    cc = 2;
             goto mvpg_progck;
-        }
+	}
 
+        /* Reset protection indication before calling translate_addr() */
+        prot = 0;
         /* Translate the first operand address to a real address */
-        rc = translate_addr (vaddr1, r1, regs, ACCTYPE_WRITE, &raddr1,
-                        &xcode, &priv, &prot, &stid, &xpblk1, &xpkey1);
+        rc = ARCH_DEP(translate_addr) (vaddr1, r1, regs, ACCTYPE_WRITE, &raddr1,
+                        &xcode, &priv, &prot, &stid);
 
-        /* If page is invalid in real storage but valid in expanded
-           storage then xpblk1 now contains expanded storage block# */
-        if (rc == 5) xpvalid1 = 1;
+	raddr1 = APPLY_PREFIXING (raddr1, regs->PX);
+
+        if (raddr1 < regs->mainsize)
+            SIE_TRANSLATE(&raddr1,ACCTYPE_READ,regs);
+
+#if defined(FEATURE_EXPANDED_STORAGE)
+        if(rc == 2)
+        {
+	    FETCH_W(pte1,sysblk.mainstor + raddr1);
+            /* If page is invalid in real storage but valid in expanded
+               storage then xpblk1 now contains expanded storage block# */
+            if(pte1 & PAGETAB_ESVALID)
+            {
+                xpblk1 = pte1 & ZPGETAB_PFRA >> 12;
+#if defined(_FEATURE_SIE)
+                if(regs->sie_state)
+                {
+                    /* Add expanded storage origin for this guest */
+                    xpblk1 += regs->sie_xso;
+                    /* If the block lies beyond this guests limit
+                       then we must terminate the instruction */
+                    if(xpblk1 >= regs->sie_xsl)
+                    {
+                        cc = 1;
+                        goto mvpg_progck;
+                    }
+                }
+#endif /*defined(_FEATURE_SIE)*/
+
+                /* If the expanded storage block does not exist
+                   then terminate the instruction */
+                if (xpblk1 >= sysblk.xpndsize)
+                {
+                    cc = 1;
+                    goto mvpg_progck;
+                }
+
+                rc = 0;
+                xpvalid1 = 1;
+                xpkey1 = sysblk.mainstor[raddr2 +
+#if defined(FEATURE_ESAME)
+                                                  2048
+#else /*!defined(FEATURE_ESAME)*/
+                /* For ESA/390 mode, the XPTE lies directly beyond 
+                   the PTE, and each entry is 12 bytes long, we must
+                   therefor add 1024 + 8 times the page index */
+	                             1024 + ((vaddr1 & 0x000FF000) >> 9)
+#endif /*!defined(FEATURE_ESAME)*/
+                                                      ];
+/*DEBUG*/ logmsg("MVPG pte1 = " F_CREG ", xkey1 = %2.2X\n",pte1,xpkey1);
+            }
+            else
+            {
+                cc = 1;
+                goto mvpg_progck;
+            }
+        }
+#endif /*defined(FEATURE_EXPANDED_STORAGE)*/
 
         /* Program check if operand not valid in main or expanded */
-        if (rc != 0 && xpvalid1 == 0)
+        if (rc)
         {
-            xaddr = vaddr1;
+	    cc = 1;
             goto mvpg_progck;
-        }
+	}
 
         /* Program check if page protection or access-list controlled
            protection applies to the first operand */
         if (prot)
         {
-            program_interrupt (regs, PGM_PROTECTION_EXCEPTION);
-            return 3;
+            regs->TEA = vaddr1 | stid | TEA_PROT_AP;
+            regs->excarid = (ACCESS_REGISTER_MODE(&regs->psw)) ? r1 : 0;
+            ARCH_DEP(program_interrupt) (regs, PGM_PROTECTION_EXCEPTION);
         }
 
     } /* end if(!REAL_MODE) */
 
+#if defined(FEATURE_EXPANDED_STORAGE)
     /* Program check if both operands are in expanded storage, or
        if first operand is in expanded storage and the destination
        reference intention (register 0 bit 22) is set to one */
     if ((xpvalid1 && xpvalid2)
-        || (xpvalid1 && (regs->gpr[0] & 0x00000200)))
+        || (xpvalid1 && (regs->GR_L(0) & 0x00000200)))
     {
-        xaddr = vaddr1;
         xcode = PGM_PAGE_TRANSLATION_EXCEPTION;
         goto mvpg_progck;
     }
@@ -379,60 +427,67 @@ BYTE    xpkey1, xpkey2;                 /* Expanded storage keys     */
     {
         /* Perform protection check on expanded storage block */
         if (akey1 != 0 && akey1 != (xpkey1 & STORKEY_KEY))
-        {
-            program_interrupt (regs, PGM_PROTECTION_EXCEPTION);
-            return 3;
-        }
+            ARCH_DEP(program_interrupt) (regs, PGM_PROTECTION_EXCEPTION);
     }
     else
+#endif /*defined(FEATURE_EXPANDED_STORAGE)*/
     {
         /* Obtain absolute address of main storage block,
            check protection, and set reference and change bits */
-        aaddr1 = logical_to_abs (vaddr1, r1, regs,
+        aaddr1 = LOGICAL_TO_ABS (vaddr1, r1, regs,
                                 ACCTYPE_WRITE, akey1);
     }
 
+#if defined(FEATURE_EXPANDED_STORAGE)
     if (xpvalid2)
     {
         /* Perform protection check on expanded storage block */
-        if (akey1 != 0 && (xpkey1 & STORKEY_FETCH)
-            && akey1 != (xpkey1 & STORKEY_KEY))
-        {
-            program_interrupt (regs, PGM_PROTECTION_EXCEPTION);
-            return 3;
-        }
+        if (akey2 != 0 && (xpkey2 & STORKEY_FETCH)
+            && akey2 != (xpkey2 & STORKEY_KEY))
+            ARCH_DEP(program_interrupt) (regs, PGM_PROTECTION_EXCEPTION);
     }
     else
+#endif /*defined(FEATURE_EXPANDED_STORAGE)*/
     {
         /* Obtain absolute address of main storage block,
-           check protection, and set reference bit */
-        aaddr2 = logical_to_abs (vaddr2, r2, regs,
+           check protection, and set reference bit. 
+           Use last byte of page to avoid FPO area.  */
+        aaddr2 = LOGICAL_TO_ABS (vaddr2 + 0xFFF, r2, regs,
                                 ACCTYPE_READ, akey2);
+        aaddr2 &= 0xFFFFF000;
     }
 
+#if defined(FEATURE_EXPANDED_STORAGE)
     /* Perform page movement */
     if (xpvalid2)
     {
         /* Set the main storage reference and change bits */
         STORAGE_KEY(aaddr1) |= (STORKEY_REF | STORKEY_CHANGE);
 
+	/* Set Expanded Storage reference bit in the PTE */
+        STORE_W(sysblk.mainstor + raddr2, pte2 | PAGETAB_ESREF);
+        
+
         /* Move 4K bytes from expanded storage to main storage */
         memcpy (sysblk.mainstor + aaddr1,
                 sysblk.xpndstor + (xpblk2 << XSTORE_PAGESHIFT),
                 XSTORE_PAGESIZE);
-        FRAG_INVALIDATE(aaddr, XSTORE_PAGESIZE);
     }
     else if (xpvalid1)
     {
         /* Set the main storage reference bit */
         STORAGE_KEY(aaddr2) |= STORKEY_REF;
 
+	/* Set Expanded Storage reference and change bits in the PTE */
+        STORE_W(sysblk.mainstor + raddr1, pte1 | PAGETAB_ESREF | PAGETAB_ESCHA);
+
         /* Move 4K bytes from main storage to expanded storage */
         memcpy (sysblk.xpndstor + (xpblk1 << XSTORE_PAGESHIFT),
                 sysblk.mainstor + aaddr2,
                 XSTORE_PAGESIZE);
     }
     else
+#endif /*defined(FEATURE_EXPANDED_STORAGE)*/
     {
         /* Set the main storage reference and change bits */
         STORAGE_KEY(aaddr1) |= (STORKEY_REF | STORKEY_CHANGE);
@@ -442,22 +497,43 @@ BYTE    xpkey1, xpkey2;                 /* Expanded storage keys     */
         memcpy (sysblk.mainstor + aaddr1,
                 sysblk.mainstor + aaddr2,
                 XSTORE_PAGESIZE);
-        FRAG_INVALIDATE(aaddr1, XSTORE_PAGESIZE);
     }
 
     /* Return condition code zero */
-    return 0;
+    regs->psw.cc = 0;
+    return;
 
 mvpg_progck:
-    /* If page translation exception and condition code option
-       in register 0 bit 23 is set, return condition code */
-    if ((regs->gpr[0] & 0x00000100)
-        && xcode == PGM_PAGE_TRANSLATION_EXCEPTION)
-        return (xaddr == vaddr2 ? 2 : 1);
+    /* If page translation exception (PTE invalid) and condition code
+        option in register 0 bit 23 is set, return condition code */
+    if ((regs->GR_L(0) & 0x00000100)
+        && xcode == PGM_PAGE_TRANSLATION_EXCEPTION
+        && rc == 2)
+    {
+        regs->psw.cc = cc;
+        return;
+    }
 
     /* Otherwise generate program check */
-    program_interrupt (regs, xcode);
-    return 3;
+    /* (Bit 29 of TEA is on for PIC 11 & operand ID also stored) */
+    if (xcode == PGM_PAGE_TRANSLATION_EXCEPTION)
+    {
+        regs->TEA |= TEA_MVPG;
+        regs->opndrid = (r1 << 4) | r2;
+    }
+    ARCH_DEP(program_interrupt) (regs, xcode);
 } /* end function move_page */
 
 #endif /*defined(FEATURE_MOVE_PAGE_FACILITY_2)*/
+
+
+#if !defined(_GEN_ARCH)
+
+#define  _GEN_ARCH 390
+#include "xstore.c"
+
+#undef   _GEN_ARCH
+#define  _GEN_ARCH 370
+#include "xstore.c"
+
+#endif /*!defined(_GEN_ARCH)*/
